// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.cmd;

import tango.io.Stdout;
import tango.io.FilePath;
import tango.io.Path;
import tango.text.Util;

import semitwist.cmdlineparser;
import semitwist.treeout;
import semitwist.util.all;

class Cmd
{
	mixin(getter!(bool, "shouldExit"));

	public this(char[][] args)
	{
		init();
		shouldExit = !parse(args);
	}
	
	// Direct command-line args (may be baked, not guaranteed to be raw)
	public bool   help;
	public bool   detailhelp;
	//public char[] engine = "gold";
	public char[] lang;
	public char[] inputFilename;

	public bool   saveParseTree=false;
	public bool   parseTreePretty=false;
	public char[] parseTreeOut;
	public char[] parseTreeType="json";

	public bool   saveTokens=false;
	public bool   tokensPretty=false;
	public char[] tokensOut;
	public char[] tokensType="json";

	public bool debugOut=false;

	// Generated
	public TreeFormatter parseFormatter;
	public TreeFormatter tokenFormatter;
	public char[] langFile;
	public char[] langSrc;
	public char[] langDir;
	public char[] langDirBuiltin;

	// Private ------
	private CmdLineParser cmd;
	private const char[] langSubDir = "lang/";
	
	private void init()
	{
		//TODO: Think about turning defineArg!() into a runtime member func of CmdLineParser
		//TODO: Search for available languages
		//TODO: Add either "quiet" or "verbose" flag
		cmd = new CmdLineParser();
		mixin(defineArg!(cmd, "help",    help,            ArgFlag.Optional, "Displays a help summary and exits" ));
		mixin(defineArg!(cmd, "detail",  detailhelp,      ArgFlag.Optional, "Displays a detailed help message and exits" ));
		//mixin(defineArg!(cmd, "engine",  engine,          ArgFlag.Optional|ArgFlag.Advanced|ArgFlag.ToLower, `Language Engine: "gold" or "stca" ("stca" disabled for now)` ));
		mixin(defineArg!(cmd, "lang",    lang,            ArgFlag.Optional|ArgFlag.Advanced,   "Language ('{lang}.cgt' and '{lang}.grmp' must exist)" ));
		mixin(defineArg!(cmd, "langdir", langDir,         ArgFlag.Optional|ArgFlag.Advanced,   "Directory to containing language files" ));
		mixin(defineArg!(cmd, "src",     inputFilename,   ArgFlag.Required|ArgFlag.Switchless, "Input filename" ));
		mixin(defineArg!(cmd, "pt",      saveParseTree,   ArgFlag.Optional, "Save Parse Tree" ));
		mixin(defineArg!(cmd, "ptp",     parseTreePretty, ArgFlag.Optional|ArgFlag.Advanced,   `Pretty Format Parse Tree File (Implies /pt)` ));
		mixin(defineArg!(cmd, "ptt",     parseTreeType,   ArgFlag.Optional|ArgFlag.Advanced|ArgFlag.ToLower, `Parse Tree File Type (Implies /pt) ("json" or "xml")` ));
		mixin(defineArg!(cmd, "pto",     parseTreeOut,    ArgFlag.Optional|ArgFlag.Advanced,   "Parse Tree Output Filename" ));
		mixin(defineArg!(cmd, "tk",      saveTokens,      ArgFlag.Optional, "Save Tokens" ));
		mixin(defineArg!(cmd, "tkp",     tokensPretty,    ArgFlag.Optional|ArgFlag.Advanced,   `Pretty Format Token File (Implies /tt)` ));
		mixin(defineArg!(cmd, "tkt",     tokensType,      ArgFlag.Optional|ArgFlag.Advanced|ArgFlag.ToLower, `Token File Type (Implies /tt) ("json" or "xml")` ));
		mixin(defineArg!(cmd, "tko",     tokensOut,       ArgFlag.Optional|ArgFlag.Advanced,   "Token Output Filename" ));
		mixin(defineArg!(cmd, "debug",   debugOut,        ArgFlag.Optional|ArgFlag.Advanced,   "Display Debug Information" ));
		
		//mixin(setArgAllowableValues!("engine", "gold", "stca"));
		mixin(setArgAllowableValues!("ptt",    "json", "xml"));
		mixin(setArgAllowableValues!("tkt",    "json", "xml"));
	}
	
	// Returns: Should processing proceed? If false, the program should exit.
	private bool parse(char[][] args)
	{
		cmd.parse(args);
		if(detailhelp)
		{
			Stdout.formatln("{}", cmd.errorMsg);
			Stdout.format("{}", cmd.getDetailedUsage());
			return false;
		}
		if(!cmd.success || help)
		{
			Stdout.formatln("{}", cmd.errorMsg);
			Stdout.format("{}", cmd.getUsage(16));
			return false;
		}
		
		if(parseTreePretty || parseTreeOut != "")
			saveParseTree = true;
		if(parseTreeOut == "")
			parseTreeOut = inputFilename~".pt."~parseTreeType;
			
		if(tokensPretty || tokensOut != "")
			saveTokens = true;
		if(tokensOut == "")
			tokensOut = inputFilename~".tk."~tokensType;
		
		parseFormatter = saveParseTree? getFormatter(parseTreeType, parseTreePretty) : null;
		tokenFormatter = saveTokens?    getFormatter(   tokensType,    tokensPretty) : null;
		
		auto pathManip = new FilePath(getExec());
		pathManip.pop();
		pathManip.append(langSubDir);
		pathManip.native();
		langDirBuiltin = pathManip.toString();
		
		if(langDir != "")
		{
			if(langDir[$-1] != '/' && langDir[$-1] != '\\')
				langDir ~= pathSeparator;
		}
		else
			langDir = langDirBuiltin;
		
		if(lang == "")
		{
			auto indexOfExt = inputFilename.locatePrior('.');
			lang = (indexOfExt<inputFilename.length) ? inputFilename[indexOfExt+1..$] : "";
		}
		
		if(lang.contains('/') || lang.contains('\\'))
		{
			Stdout.formatln("Value for '/lang:' must not contain a path.");
			Stdout.formatln("To specify a path to the language files, use '/langdir:'.");
			Stdout.formatln("");
			Stdout.format("{}", cmd.getUsage(16));
			return false;
		}
		
		langFile = langDir~lang~".cgt";
		langSrc  = langDir~lang~".grmp";
		if(!exists(langSrc))
		{
			char[] langSrcAlt = langDir~lang~".grm";
			if(exists(langSrcAlt))
				langSrc = langSrcAlt;
		}
		
		return true;
	}
	
	private TreeFormatter getFormatter(char[] type, bool pretty)
	{
		TreeFormatter formatter = 
			(type=="xml"  &&  pretty)? formatterPrettyXML   :
			(type=="xml"  && !pretty)? formatterTrimmedXML  :
			(type=="json" &&  pretty)? formatterPrettyJSON  :
			(type=="json" && !pretty)? formatterTrimmedJSON :
			null;
		return formatter;
	}
}
