// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)
*/

module stc.gold;

import tango.io.Stdout;
import tango.text.Regex;
import tango.text.Util;

import semitwist.util.all;
import semitwist.treeout;

import stc.util;
import stc.language;

import stc.goldbase;
import stc.goldfile;

import tango.core.Traits;
import tango.core.Tuple;

public class TokGold : Token
{
	private GoldSymbol symbol;
	mixin(getter!(char[], "debugInfo"));
	mixin(getter!(CommentType, "commentMode"));
	mixin(getter!(GoldSymbolType, "symbolType"));
	
	mixin(getterLazy!(int, "line"));
	public int _line_gen()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().line();
		else
			return _line;
	}
	
	mixin(getterLazy!(char[], "file"));
	public char[] _file_gen()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().file();
		else
			return _file;
	}
	
	mixin(getterLazy!(int, "srcIndexStart"));
	public int _srcIndexStart_gen()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getFirstLeaf().srcIndexStart();
		else
			return _srcIndexStart;
	}
	
	mixin(getterLazy!(int, "srcIndexEnd"));
	public int _srcIndexEnd_gen()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
			return getLastLeaf().srcIndexEnd();
		else
			return _srcIndexEnd;
	}
	
	//TODO? Expose as opIndex
	private TokGold[] subTokens;
	
	public this(GoldSymbol symbol, schar[] content,
	            char[] file="{unknown}", int line=1,
				int srcIndexStart=0, int srcIndexEnd=0,
				CommentType commentMode=CommentType.None,
				char[] debugInfo="")
	{
		super(content);
		mixin(initMember!(symbol, debugInfo/*, commentMode*/));
		
		// Workaround for DMD Bug #2881
		this.commentMode = commentMode;
		
		//TODO: Come up with way to do initMember on these
		this._file = file;
		this._line = line;
		this._srcIndexStart = srcIndexStart;
		this._srcIndexEnd = srcIndexEnd;
		
		this.symbolType = cast(GoldSymbolType)symbol.type;
		
		if(symbolType == GoldSymbolType.NonTerminal)
			throw new Exception("Called wrong constructor for NonTerminal");
	}
	
	public this(GoldSymbol symbol, TokGold[] subTokens)
	{
		super("");
		this.symbolType = cast(GoldSymbolType)symbol.type;
		mixin(initMember!(symbol, subTokens));
		
		if(symbolType != GoldSymbolType.NonTerminal)
			throw new Exception("Called wrong constructor for token that isn't a NonTerminal");
	}
	
	//TODO? Convert the rest of these to getterLazy
	public TokGold getFirstLeaf()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
		{
			foreach(TokGold subToken; subTokens)
			{
				if(subToken.symbolType() != GoldSymbolType.NonTerminal || subToken.subTokens.length > 0)
					return subToken.getFirstLeaf();
			}
		}

		return this;
	}
	
	public TokGold getLastLeaf()
	{
		if(symbolType == GoldSymbolType.NonTerminal && subTokens.length > 0)
		{
			foreach_reverse(TokGold subToken; subTokens)
			{
				if(subToken.symbolType() != GoldSymbolType.NonTerminal || subToken.subTokens.length > 0)
					return subToken.getLastLeaf();
			}
		}

		return this;
	}
	
/*	public int srcLength()
	{
		return srcIndexEnd() - srcIndexStart();
	}
*/	

	public char[] symbolTypeName()
	{
		return goldSymbolTypeToString(symbolType);
	}

	public wchar[] symbolName()
	{
		return symbol.name.dup;
	}

	public wchar[] symbolFullName()
	{
		return "{}.{}"w.stformat(symbolTypeName(), symbolName());
	}
	
	public int symbolId()
	{
		return symbol.id;
	}

	//TODO: Can I do toString(bool raw=false)? (Ie, will Stdout.formatln("{}", this) still work?)
	public char[] toString()
	{
		char[] ret;
		
		switch(symbolType)
		{
		case GoldSymbolType.NonTerminal:
			foreach(TokGold tok; subTokens)
				ret ~= tok.toString();
			break;
			
		case GoldSymbolType.Terminal:
			ret = content.dup;
			break;
			
		case GoldSymbolType.Whitespace:
		case GoldSymbolType.EOF:
		case GoldSymbolType.CommentStart:
		case GoldSymbolType.CommentEnd:
		case GoldSymbolType.CommentLine:
		case GoldSymbolType.Error:
			ret = "";
			break;
			
		default:
			throw new Exception("Unknown GoldSymbolType: #{}".stformat(symbolType));
		}
		
		return ret;
	}
	
	public char[] toStringRaw()
	{
		char[] ret;
		
		if(symbolType == GoldSymbolType.NonTerminal)
		{
			foreach(TokGold tok; subTokens)
				ret ~= tok.toString();
		}
		else
			ret = content.dup;
			
		return ret;
	}
	
	public TreeNode toTreeNode(size_t index=0)
	{
		auto start = srcIndexStart();
		auto node = new TreeNode( "{}: {}".stformat(index, symbolName()) );
		node.addAttribute("srcIndexStart", start);
		//node.addAttribute("srcIndexEnd", srcIndexEnd());
		node.addAttribute("srcLength", srcIndexEnd() - start);
		node.addAttribute("line", line());
		node.addAttribute("type", symbolTypeName());
		node.addAttribute("symbol", symbolName());
		if(debugInfo() != "")
			node.addAttribute("debugInfo", debugInfo());
		//node.addAttribute("content", toStringRaw());
		
		if(symbolType == GoldSymbolType.NonTerminal)
		foreach(size_t i, TokGold tok; subTokens)
			node.addContent(tok.toTreeNode(i));
		
		return node;
	}
	
	public TreeNode toPsuedoTreeNode()
	{
		auto node = new TreeNode("{}".stformat(toStringRaw()));
		node.addAttribute("symbol", symbolFullName());

		if(symbolType == GoldSymbolType.NonTerminal)
		foreach(TokGold tok; subTokens)
			node.addContent(tok.toPsuedoTreeNode());
		
		return node;
	}
}

public class GoldLanguage : Language
{
	//TODO: Think about turning these into accessors,
	//      or just move the lexing/parsing into this class
	public wchar[] grammarName;
	public wchar[] grammarVersion;
	public wchar[] grammarAuthor;
	public wchar[] about;
	public bool    caseSensitive;
	
	public GoldSymbol[]    symbolTable;
	public GoldCharSet[]   characterSetTable;
	public GoldRule[]      ruleTable;
	public GoldDFAState[]  dfaTable;
	public GoldLALRState[] lalrTable;
	
	public int startSymbolIndex;
	public int initialDFAState;
	public int initialLALRState;
	
	public int eofSymbolIndex;
	public int errorSymbolIndex;

	public this(char[] cgtFilename, bool debugOutput=false)
	{
		auto goldFile = new GoldFile(cgtFilename);
		
		mixin(initMemberFrom!(goldFile,
		                      grammarName, grammarVersion, grammarAuthor,
							  about, caseSensitive,
		                      startSymbolIndex, initialDFAState, initialLALRState,
	                          symbolTable, characterSetTable, ruleTable,
						      dfaTable, lalrTable));

		eofSymbolIndex   = getSymbolIndexByType(GoldSymbolType.EOF);
		errorSymbolIndex = getSymbolIndexByType(GoldSymbolType.Error);

		if(debugOutput)
		{
			mixin(traceVal!("grammarName", "grammarVersion", "grammarAuthor",
							"about", "caseSensitive",
							"startSymbolIndex", "initialDFAState", "initialLALRState", 
							"symbolTable.length", "characterSetTable.length",
							"ruleTable.length", "dfaTable.length", "lalrTable.length"
							));
			mixin(traceVal!("eofSymbolIndex", "errorSymbolIndex"));
			mixin(traceVal!("symbolTable[eofSymbolIndex]", "symbolTable[errorSymbolIndex]"));
			mixin(traceVal!("symbolTable"));

			Stdout.formatln("");
			mixin(traceVal!("symbolTable", "characterSetTable",
							"ruleTable", "dfaTable", "lalrTable"
							));
								
		}
	}

	private int getSymbolIndexByType(GoldSymbolType type)
	{
		foreach(int i, GoldSymbol symbol; symbolTable)
			if(symbol.type == type)
				return i;
		throw new Exception("No symbol found by getSymbolIndexByType for type {}".stformat(type));
	}
	
	public GoldSymbol eofSymbol()
	{
		return symbolTable[eofSymbolIndex];
	}
	
	public GoldSymbol errorSymbol()
	{
		return symbolTable[errorSymbolIndex];
	}
	
	public TokenType[] getTokenTypes()
	{
		TokenType[] tokTypes;
		
		tokTypes ~= new TokenType(r":==",
						delegate Token(schar[] data)
						{
							throw new Exception("(!!!{}!!!USED TOKENTYPE!!!)".stformat(data));
						});

		return tokTypes;
	}
}
