// SemiTwist Compiler
// Written in the D programming language.

/** 
Author:
$(WEB www.semitwist.com, Nick Sabalausky)

Requires:
- DMD 1.041
- Tango 0.99.8
- semitwistlib r39
*/

module stc.main;

import tango.core.Exception;
import tango.io.Stdout;
import tango.io.device.File;
import tango.text.Util;

import semitwist.util.all;
import semitwist.treeout;

import stc.util;
import stc.cmd;
import stc.lexer;
import stc.parser;
import stc.language;
import stc.gold;

import stc.goldbase;
import stc.goldfile;

char[] appVersion = "v0.003";

enum Verbosity
{
	Low,
	Medium,
	High,
}

int main(char[][] args)
{
	flushAsserts(); // For unittests
	outputHeader(); //TODO: Only output header on help display when in quiet/non-verbose mode
	
	auto cmd = new Cmd(args);
	if(cmd.shouldExit())
		return 1;

	if(cmd.debugOut)
	{
		Stdout.formatln("Input File: {}", cmd.inputFilename);
		Stdout.formatln("Language: {}", cmd.lang);
		Stdout.formatln("Language File: {}", cmd.langFile);
		Stdout.formatln("Language Source: {}", cmd.langSrc);
	}

	auto goldLang = new GoldLanguage(cmd.langFile, cmd.debugOut);

	auto parseTree = process(cmd.inputFilename, goldLang,
							 cmd.parseTreeOut, cmd.tokensOut,
							 cmd.parseFormatter, cmd.tokenFormatter,
							 cmd.debugOut? Verbosity.High : Verbosity.Medium);
	if(parseTree)
	{
		char[] goldPlusFilename = "grmp.cgt";
		outputSection("Getting Ready to Generate AST");
		Stdout.format("Loading Gold Plus...");
		auto lang = new GoldLanguage(cmd.langDirBuiltin~goldPlusFilename, Verbosity.Low);
		Stdout.formatln("Done!");
		
		Stdout.format("Loading '{}' Language Source...", cmd.lang);
		auto grmParseTree = process(cmd.langSrc, lang,
									"", "",
									null, null,
									Verbosity.Low);
		Stdout.formatln("Done!");
		
		//TODO: Associate 'grmParseTree' with 'lang' and apply to parseTree to obtain AST
	}
	
	return 0;
}

char[] genTreeStr(TokGold rootTok, TreeFormatter formatter, char[] inputFilename, schar[] source)
{
	auto parseTreeRoot = rootTok.toTreeNode();
	parseTreeRoot.addAttribute("file", inputFilename);
	parseTreeRoot.addAttribute("parseTreeMode", "true");
	parseTreeRoot.addAttribute("source", source);
	return parseTreeRoot
	       .format(formatter)
	       .toNativeEOLFromUnix();
}

char[] genTokenTreeStr(TokGold[] tokens, TreeFormatter formatter, char[] inputFilename, schar[] source)
{
	return genTreeStr(
		new TokGold(
			GoldSymbol("Tokens", GoldSymbolType.NonTerminal, -1),
			tokens
		),
		formatter, inputFilename, source
	);
}

// T must be either TokGold or TokGold[]
void saveTree(T)(T tokens, TreeFormatter formatter, char[] label,
                 char[] outFilename, char[] inputFilename, schar[] source)
{
	if(formatter)
	{
		Stdout.format("Saving {} ({})...", label, outFilename);
		Stdout.flush();
		
		char[] treeStr;
		static if(is(T:TokGold))
			treeStr = genTreeStr(tokens, formatter, inputFilename, source);
		else static if(is(T:TokGold[]))
			treeStr = tokens.genTokenTreeStr(formatter, inputFilename, source);
		else
			static assert(false, "'T tokens' is type '"~T.stringof~"', but must be either type 'TokGold' or 'TokGold[]'");
			
		saveFile(outFilename, cast(ubyte[])treeStr);
		Stdout.formatln("Done!");
	}
}

// Returns: Root 'Token' of parse tree, or null if failed to parse.
Token process(char[] inputFilename, GoldLanguage lang,
              char[] parseTreeOut, char[] tokensOut, 
              TreeFormatter parseFormatter, TreeFormatter tokenFormatter,
              Verbosity verbosity)
{
	TokGold parseTree = null;
	auto source = loadFile(inputFilename);

	if(verbosity == Verbosity.High)
		outputSection("Input", source);
	
	TokGold[] tokens = source.lex(inputFilename, lang,
                                  tokenFormatter, tokensOut,
                                  verbosity);
	if(tokens)
	{
		parseTree = tokens.parse(source, inputFilename, lang,
                                 parseFormatter, parseTreeOut,
                                 verbosity);
	}

	return parseTree;
}

TokGold[] lex(schar[] source, char[] inputFilename, GoldLanguage lang,
              TreeFormatter formatter, char[] treeOut, Verbosity verbosity)
{
	if(verbosity != Verbosity.Low)
		outputSection("Lexing");
		
	auto lexer = new Lexer();
	TokGold[] tokens;
	try
		tokens = lexer.process(source, inputFilename, lang);
	catch(Exception e)
	//TODO: How the hell do I get the typeof(e)@module(line) like default handler? (test with assert failure)
		Stdout.formatln("Lexer: {}", e.toString());

	displayLexResults(lexer, tokens, verbosity);
	saveTree(tokens, formatter, "Tokens", treeOut, inputFilename, source);
	
	if(lexer.numErrors != 0)
		tokens = null;
	
	return tokens;
}

TokGold parse(TokGold[] tokens,
              schar[] source, char[] inputFilename, GoldLanguage lang,
              TreeFormatter formatter, char[] treeOut, Verbosity verbosity)
{
	TokGold parseTree = null;
	
	if(verbosity != Verbosity.Low)
		outputSection("Parsing");
		
	auto parser = new Parser();
	try
		tokens = parser.process(tokens, inputFilename, lang);
	catch(Exception e)
	{
		Stdout.formatln("Parser: {}", e);

		outputSection("Parser Info");
		displayTokens(parser.tokenStack);
		mixin(traceVal!("parser.stateStack", "parser.rulesUsed"));
	}
	
	if(verbosity == Verbosity.High)
	{
		outputSection("Tokens");
		displayTokens(tokens);
	}
	
	if(tokens.length == 1)
	{
		if(verbosity != Verbosity.Low)
			Stdout.formatln("Success!");
			
		parseTree = tokens[0];
		saveTree(parseTree, formatter, "Parse Tree", treeOut, inputFilename, source);
	}
	else
	{
		if(verbosity != Verbosity.Low)
			Stdout.formatln("Source Failed to Parse!");
			
		saveTree(tokens, formatter, "Parse Tree", treeOut, inputFilename, source);
	}
	
	return parseTree;
}

void displayLexResults(Lexer lexer, Token[] tokens, Verbosity verbosity)
{
	if(verbosity != Verbosity.Low)
	{
		if(lexer.numErrors != 0)
			Stdout.format("{}", lexer.errorMsg);

		Stdout.formatln("{} error{}",
		                lexer.numErrors,
		                lexer.numErrors == 1 ? "" : "s");
	
		if(verbosity == Verbosity.High)
		{
			outputSection("Remaining Data", lexer.remainingData.toNativeEOLFromUnix());
			outputSection("Processed Data", lexer.processedData.toNativeEOLFromUnix());
			outputSection("Mixed Data",     lexer.mixedData.toNativeEOLFromUnix());
			outputSection("Tokens");

			displayTokens(tokens);
		}

		if(lexer.remainingData == "")
			Stdout.formatln("Success!");
	}
}

void displayTokens(Token[] tokens)
{
	bool gold=false;
	foreach(int i, Token tok; tokens)
	{
		TokGold tokGold = cast(TokGold)tok;
		if(tokGold)
		{
			gold = true;
			if(tokGold.symbolType() != GoldSymbolType.Whitespace)
				Stdout.formatln("{}({}): [{}]{,-25} '{}'",
								tokGold.file(), tokGold.line(), i,
								tokGold.symbolFullName() ~ ":",
								tokGold.toStringRaw());

		}
	}
}

void outputHeader()
{
	//Stdout.formatln("");
	Stdout.formatln("SemiTwist Compiler {}", appVersion);
	Stdout.formatln("");
}

void outputSection(T,U=void)(T[] heading, U[] content=null)
{
	//Stdout.formatln("");
	Stdout.formatln("------- {}:", heading);
	if(content != null)
		Stdout.formatln("{}", content);
}

schar[] loadFile(char[] filename)
{
	auto inFile = new File(filename);
	auto inFileLength = inFile.length;
	//WARNING: file can't be larger than uint.max
	//TODO?: Convert lex/parse to work as producers/consumers, like in GOLD docs,
	//       instead of all fully completing each step before next.
	schar[] inFileContent = cast(schar[])(new byte[cast(uint)inFileLength]);
	auto inFileLengthRead = inFile.input.read(inFileContent);
	inFile.close();
	
	if(inFileLengthRead != inFileLength)
		throw new Exception(
			"Failed to read entire file: {}: Read {} out of {} bytes."
				.stformat(filename, inFileLengthRead, inFileLength)
		);

	return inFileContent;
}

void saveFile(char[] filename, ubyte[] data)
{
	auto outFile = new File(filename, File.WriteCreate);
	auto bytesWritten = outFile.output.write(data);
	outFile.close();
	
	if(bytesWritten != data.length)
		throw new Exception(
			"Failed to write entire file: {}: Wrote {} out of {} bytes."
				.stformat(filename, bytesWritten, data.length)
		);
}
